#ifndef PN_TRANSITION_H
#define PN_TRANSITION_H

#include <list>
#include "PN_Condition.h"
#include "Bindings.h"
#include "PN_Exception.h"

using namespace std;

class PN_Expression;
class PN_Action;
class PN_Guard;

class PN_Checkable_Port
{
protected:
    PN_Guard *guard;
    list<PN_Condition *> preconds;
    list<PN_Condition *> conds;
    list<PN_Condition *> postconds;

public:
    string name;
/***
  * @param init_binds : intended for use within synchronous port to pass binded variables
  *
  */
    virtual PN_TransitionBinding *GetBinding(PN_TransitionBinding *init_binds = NULL);

    ~PN_Checkable_Port();
    void SetPreconds(list<PN_Condition *> preC);
    void SetPostconds(list<PN_Condition *> preC);
    void SetConds(list<PN_Condition *> preC);
    void SetGuard (PN_Guard *g);
};

class PN_Transition : public PN_Checkable_Port
{
public:
    PN_Transition(string name);
    virtual ~PN_Transition();

    void ExecuteWithBinding(PN_TransitionBinding *binding);
    void AddAction (PN_Expression *act);

    void Die();
    bool IsDead();
private:
    list<PN_Expression *> actions;
    bool died;
public:
    //string name;
};

class PN_Action
{
private:
    list<PN_Expression*>::iterator act_expr;
    list<PN_Expression*> expr;
    list<PN_Binding> actionNamespace;
    list<PN_Condition *> postconds;
public :
    PN_Action(list<PN_Expression*> _expr, list<PN_Binding> binds, list<PN_Condition *> _postconds)
        : expr(_expr), actionNamespace(binds), postconds(_postconds)
    {
        act_expr = expr.begin();
    }
    bool Execute();
    bool Continue(Token *jMsgResult);
};

class PN_Guard
{
    list<PN_Expression*> expr;
    PN_Guard *nextGuard;
    bool Execute(list<PN_Binding>& actionNamespace, bool simulateOnly);
public:
    PN_Guard() {nextGuard = NULL;}
    ~PN_Guard() {DisposeRefList(this->expr); delete nextGuard;}
    void AddExpr(PN_Expression* e) {expr.push_back(e);}
    void SetGuard (PN_Guard *next) {nextGuard = next;}
    bool CheckGuard(list<PN_Binding>& actionNamespace);
    bool ExecuteGuard(list<PN_Binding>& actionNamespace);
};

class PN_JoinException
{
private:
    PN_Action* joinTarget;
    Token* joinResult;
public:
    PN_JoinException(PN_Action* target, Token *result)
    {
        joinTarget = target;
        joinResult = result;
    }
    virtual void Join()
    {
        bool actionFinished = false;
        try
        {
            actionFinished = joinTarget->Continue(joinResult);
        }
        catch(PN_JoinException *ex)
        {
            ex->Join();
            delete ex;
        }
        if(actionFinished)
        {
            delete joinTarget;
        }
    }
    virtual ~PN_JoinException()
    {

    }
};

class PN_ForkException : public PN_Exception
{
public:
    virtual void SetActionInstance(PN_Action* contextAction)=0;
    virtual void CancelFork()=0;
};



class PN_Message
{
protected:
    string text;
public:
    PN_Message(string msg)
    {
        this->text = msg;
    }
    virtual ~PN_Message() {}
    virtual Token *ExecuteAction(Token *op, list<PN_Binding>& args, bool simulateOnly)=0;
};

class PN_MessageUnary : public PN_Message
{
public:
    PN_MessageUnary(string msg) : PN_Message(msg) {}
    virtual ~PN_MessageUnary() {}
    virtual Token *ExecuteAction(Token *op, list<PN_Binding>& args, bool simulateOnly);
};


class PN_MessageBin : public PN_Message
{
public:
    PN_MessageBin(string msg, PN_Operand* op) : PN_Message(msg) , operand(op) {}
    virtual ~PN_MessageBin() { delete operand;}
    PN_Operand *operand;
    virtual Token *ExecuteAction(Token *op, list<PN_Binding>& args, bool simulateOnly);
};


class PN_MessageKeys : public PN_Message
{
public:
    PN_MessageKeys(string msg) : PN_Message(msg) {}
    virtual ~PN_MessageKeys() { DisposeRefList(keys);}
    list<PN_Operand*> keys;
    virtual Token *ExecuteAction(Token *op, list<PN_Binding>& args, bool simulateOnly);
};


class PN_Expression
{
private:
    PN_Message *msg;
    PN_Operand *rootOperand;
    list<string> resultVarname;
public:
    PN_Expression(PN_Message *_msg, PN_Operand *target) : msg(_msg), rootOperand(target) {}
    virtual ~PN_Expression() { delete msg; delete rootOperand; }

    void UpdateNamespace(list<PN_Binding>& args, Token* value);
    Token *ExecuteAction(list<PN_Binding>& args, bool simulateOnly);

    void AddResultVar(string var) { resultVarname.push_back(var);}

};


#endif // PN_TRANSITION_H
