#ifndef PETRINET_H
#define PETRINET_H

#include <list>
#include "PN_Transition.h"

using namespace std;

class PN_MetaClass;
class PN_Class;

/** @brief base class for all Petri Nets
  *
  * Base class for elements with petri net. Maintains transitions for net instance.
  */
class PN
{
public:
    PN();
    virtual ~PN();
    virtual PN_Place *GetPlace(string placeName) {return NULL;}
protected:
    void RegisterTransition(PN_Transition *t);
    virtual PN_Class *GetSelf() = 0;
private:
    list<PN_Transition*> transitions;
};

class PN_SyncPort : public PN_Checkable_Port
{
private:
    list<string> argNames;
public:
    PN_SyncPort(string name) {this->name = name; this->guard = NULL;}
    PN_SyncPort(list<string> args, string name);
    bool TryBind(list<PN_Operand*> ops, list<PN_Binding>& binds);
    bool Execute(list<PN_Operand*> ops, list<PN_Binding>& binds);
    string name;
};

/** @brief PNTalk class base
  *
  * Base Class providing message handling
  */
class PN_Class : public PN
{
protected:
    PN_Class* GetSelf() {return this;}
public:
    virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual Token *SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual string AsText()=0;

};

/** @brief token carrying PNTalk class instance
  *
  * Net is accessed as reference, there may be more tokens referencing same instance
  */
class PNToken : public ReferenceToken
{
public:
    PNToken(PN_Class *_value) : value (_value) {}
    Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    Token *SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual int GetIntertokenOrder() { return 6; }
    virtual void Print() { cout << "PN object token(" << typeid(*value).name()<< ")"; }
    virtual string AsText() {ostringstream text; text << "PN object token(" << typeid(*value).name()<< ")"; return text.str();}
    virtual void* GetValue() { return value; }
    virtual Token* Copy() {return new PNToken(value);}
    PN_Class* value;
};

class PN_MetaClass : public ReferenceToken
{
protected:
    virtual PNToken *GetNewInstance()=0;
public:
    Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
    virtual void* GetValue() { return NULL; }
    virtual string AsText() { return "Meta class"; }
    virtual Token* Copy() {return NULL;}
};

class PN_Method_Net : public PN
{
    friend class PN_Place_Return;
protected:
    PN_Action *caller;
    PN_Class *container;
    PN_Class* GetSelf() {return this->container;}
public:
    PN_Method_Net(PN_Class *parentObj) {container = parentObj;}
    void SetAction (PN_Action* action) { caller = action; }
    virtual PN_Place *GetPlace(string placeName) {return this->container->GetPlace(placeName);}
};

class PN_MethodForkException: public PN_ForkException
{
    private:
        PN_Method_Net *methodInstance;
    public:
        PN_MethodForkException(PN_Method_Net* m) {methodInstance = m;}
        void SetActionInstance(PN_Action* contextAction)
        {
            if (methodInstance != NULL)
                methodInstance->SetAction(contextAction);
            else
                throw PN_Error_Exception("cannot set action to forked method net after cancel called.");
        }
        void CancelFork() { delete methodInstance; methodInstance = NULL;}
};

/** @brief special place for returning from method net
  *
  */
class PN_Place_Return : public PN_Place
{
public:
    PN_Place_Return(PN_Method_Net *_parent) : PN_Place("return"), parent(_parent) {}
    virtual void PushTokens(list<Token*> *tokens);
private:
    PN_Method_Net *parent;
};

class PN_JoinMethodException : public PN_JoinException
{
    private:
        PN_Method_Net *methodInstance;
    public:
        PN_JoinMethodException(PN_Action* target, Token *result, PN_Method_Net *instance)
            : PN_JoinException(target, result)
            { methodInstance = instance ;}
        virtual void Join() {delete methodInstance; PN_JoinException::Join(); }
};


#endif // PETRINET_H
