#ifndef TOKEN_H
#define TOKEN_H

#include <string>
#include <iostream>
#include <typeinfo>
#include <list>
#include <vector>
#include <sstream>
#include "PN_Exception.h"

using namespace std;

template< typename T>
void DisposeRefList (list<T>& l)
{
    typename list<T> ::iterator it_var, it_end;
    it_var = l.begin();
    it_end = l.end();

    for(; it_var != it_end; it_var++) //remove every item from multiset
    {
        delete *it_var;
    }
};


template< typename T>
void DisposeRefList (vector<T>& l)
{
    typename vector<T> ::iterator it_var, it_end;
    it_var = l.begin();
    it_end = l.end();

    for(; it_var != it_end; it_var++) //remove every item from multiset
    {
        delete *it_var;
    }
}



typedef enum {nil_token, integer_token, bool_token, character_token, string_token, symbol_token, reference_token, list_token} TokenType;

class Token;
class PN_Binding;

Token* FindBinding(list<PN_Binding> *partialBinding, string varName);

class PN_Operand
{
public:
    PN_Operand() {}
    virtual ~PN_Operand() {}
    virtual Token *Resolve(list<PN_Binding> args)=0;
};

class Token
{
    public:
        Token();
        virtual ~Token();
        virtual Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual Token *SimulateMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder()=0;
        virtual int Compare (Token *other);
        virtual string AsText() {return "Base token";}
        virtual void Print() {cout << this->AsText();}
        virtual Token* Copy()=0;
        virtual TokenType GetType()=0;
    protected:
    private:
};



class PN_Operand_Var: public PN_Operand
{
private:
    string varName;
public:
    PN_Operand_Var(string var)
    {
        varName = var;
    }
    virtual ~PN_Operand_Var() {}
    virtual Token *Resolve(list<PN_Binding> args);
};

class PN_Operand_Token: public PN_Operand
{
private:
    Token *bindedToken;
public:
    PN_Operand_Token(Token *token)
    {
        bindedToken = token;
    }
    virtual ~PN_Operand_Token() {delete bindedToken; bindedToken = NULL;}
    Token *Resolve(list<PN_Binding> args);
};

class PN_Binding
{
    public:
        PN_Binding (const PN_Binding& obj) ;
        PN_Binding (string var, Token* trg) {varName = var; target = trg;}
        string varName;
        Token *target;
        string AsText() {ostringstream text; text << varName << " = " << target->AsText(); return text.str();}
};

class PN_MessagePass_Exception : public PN_Error_Exception
{
public:
    PN_MessagePass_Exception(string msg) : PN_Error_Exception(msg) {}
};

class PN_MessageNotUnderstood_Exception: public PN_MessagePass_Exception
{
    private:
        string message;
        string target;
    public:
        PN_MessageNotUnderstood_Exception (string _target, string msg)
            : PN_MessagePass_Exception(""), message(msg), target(_target) {}
        virtual void Print()
        {
            PN_Exception::Print();
            cout << ": token " << target;
            cout << "did not understand message \"" << this->message << "\"\n";
        }
};

class PN_CantResolve_Exception : public PN_MessagePass_Exception
{
public:
    string var;
    PN_CantResolve_Exception(string varName) : PN_MessagePass_Exception("")
    {
        var = varName;
        ostringstream s;
        s << "no binding to variable " << varName << " found";
        this->msg = s.str();
    }
};

class PN_NotEnoughArguments : public PN_MessagePass_Exception
{
private:
    string message;
    string className;
    int requiredArgs;
    int providedArgs;
public:
    PN_NotEnoughArguments(string _target, string msg, int required, int provided)
        : PN_MessagePass_Exception(""), message(msg), className(_target),
        requiredArgs(required), providedArgs(provided) {}
    virtual void Print()
    {
        PN_Exception::Print();
        cout << ": class of type "<< className << ", message "<< message<<"- insufficient count of arguments ("
        <<providedArgs << " provided, "<< requiredArgs<< " reqired)" << endl;
    }


};

bool operator < (Token& lhs, Token& rhs);
bool operator == (Token& lhs, Token& rhs);

bool compare_tokens_lt (Token* lhs, Token* rhs);
int compare_tokens_cmp (Token* lhs, Token* rhs);
int compare_lists_cmp (list<Token*>& lhs, list<Token*>& rhs);

class IntToken : public Token
{
    public:
        IntToken(int _value);
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 1;}
        virtual string AsText() {stringstream t; t << value; return t.str();}
        virtual int Compare(Token *other);
        virtual Token* Copy() {return new IntToken(value);}
        virtual TokenType GetType() {return integer_token;}
        int value;
};

class BoolToken : public Token
{
    public:
        BoolToken(bool _value) : value (_value) {}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 2;}
        virtual string AsText() {return value? "true" : "false";}
        virtual int Compare(Token *other);
        virtual Token* Copy() {return new BoolToken(value);}
        virtual TokenType GetType() {return bool_token;}
        bool value;
};

class CharToken : public Token
{
    public:
        CharToken(char _value) : value (_value) {}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 3;}
        virtual string AsText() {stringstream t; t << "$" << value; return t.str();}
        virtual int Compare(Token *other);
        virtual Token* Copy() {return new CharToken(value);}
        virtual TokenType GetType() {return character_token;}
        char value;
};

class StringToken : public Token
{
    public:
        StringToken(string _value) : value (_value) {}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 4;}
        virtual string AsText() {stringstream t; t << "'" << value << "'"; return t.str();}
        virtual int Compare(Token *other);
        virtual Token* Copy() {return new StringToken(value);}
        virtual TokenType GetType() {return string_token;}
        string value;
};

class SymbolToken : public Token
{
    public:
        SymbolToken(string _value) : value (_value) {}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 5;}
        virtual void Print() {cout << "#" << value;}
        virtual string AsText() {stringstream t; t << "#" << value; return t.str();}
        virtual int Compare(Token *other);
        virtual Token* Copy() {return new SymbolToken(value);}
        virtual TokenType GetType() {return symbol_token;}
        string value;
};

class ReferenceToken : public Token
{
    public:
        ReferenceToken() {}
        virtual int GetIntertokenOrder() {return 6;}
        virtual int Compare(Token *other) ;
        TokenType GetType() {return reference_token;}
        virtual void* GetValue()=0;
};

class ListToken : public Token
{
    public:
        ListToken() {}
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds);
        virtual int GetIntertokenOrder() {return 7;}
        virtual string AsText();
        virtual int Compare(Token *other);
        virtual Token* Copy()
        {
            ListToken *result = new ListToken();

            list<Token*>::iterator it_var, it_end;
            it_var = this->collection.begin();
            it_end = this->collection.end();

            for(; it_var != it_end; it_var++)
            {
                result->collection.push_back((*it_var)->Copy());
            }
            return result;
        }
        virtual TokenType GetType() {return list_token;}
        list<Token*> collection;

};

class NilToken : public Token
{
    public:
        Token *SendMessage(string message, list<PN_Operand*>& args, list<PN_Binding>& binds)
            {throw new PN_MessageNotUnderstood_Exception(this->AsText(), message);}
        virtual int GetIntertokenOrder() {return -2;}
        string AsText() {return "nil";}
        virtual TokenType GetType() {return nil_token;}
        virtual Token* Copy() {return new NilToken();}
};



#endif // TOKEN_H
